"""
lex
~~~

A lexer for `yadr` dice notation.
"""
from functools import wraps
from typing import Callable, Optional

from yadr import maps
from yadr.base import BaseLexer
from yadr.model import (
    BaseToken,
    CompoundResult,
    Result,
    Token,
    TokenInfo,
    symbols
)


# Lexers.
class Lexer(BaseLexer):
    """A state-machine to lex dice notation."""
    def __init__(self) -> None:
        state_map: dict[BaseToken, Callable] = {
            Token.START: self._start,
            Token.AS_OPERATOR: self._as_operator,
            Token.BOOLEAN: self._boolean,
            Token.CHOICE_OPERATOR: self._choice_operator,
            Token.COMPARISON_OPERATOR: self._comparison_operator,
            Token.DICE_OPERATOR: self._dice_operator,
            Token.EX_OPERATOR: self._ex_operator,
            Token.GROUP_OPEN: self._group_open,
            Token.GROUP_CLOSE: self._group_close,
            Token.MAP: self._map,
            Token.MAP_END: self._map_end,
            Token.MAPPING_OPERATOR: self._mapping_operator,
            Token.MD_OPERATOR: self._md_operator,
            Token.NUMBER: self._number,
            Token.OPTIONS_OPERATOR: self._options_operator,
            Token.POOL: self._pool,
            Token.POOL_DEGEN_OPERATOR: self._pool_degen_operator,
            Token.POOL_END: self._pool_end,
            Token.POOL_GEN_OPERATOR: self._pool_gen_operator,
            Token.POOL_OPERATOR: self._pool_operator,
            Token.QUALIFIER: self._qualifier,
            Token.QUALIFIER_END: self._qualifier_end,
            Token.ROLL_DELIMITER: self._roll_delimiter,
            Token.U_POOL_DEGEN_OPERATOR: self._u_pool_degen_operator,
            Token.WHITESPACE: self._whitespace,
            Token.END: self._start,
        }
        bracket_states: dict[BaseToken, BaseToken] = {
            Token.MAP_OPEN: Token.MAP,
            Token.NEGATIVE_SIGN: Token.NUMBER,
            Token.QUALIFIER_DELIMITER: Token.QUALIFIER,
            Token.POOL_OPEN: Token.POOL,
        }
        bracket_ends: dict[BaseToken, BaseToken] = {
            Token.MAP: Token.MAP_END,
            Token.QUALIFIER: Token.QUALIFIER_END,
            Token.POOL: Token.POOL_END,
        }
        symbol_map: dict[BaseToken, list[str]] = symbols
        result_map: dict[BaseToken, Callable] = {
            Token.BOOLEAN: self._tf_boolean,
            Token.MAP: self._tf_maps,
            Token.NUMBER: self._tf_number,
            Token.POOL: self._tf_pool,
            Token.QUALIFIER: self._tf_qualifier,
        }
        no_store: list[BaseToken] = [
            Token.WHITESPACE,
            Token.START,
            Token.MAP_END,
            Token.POOL_END,
            Token.QUALIFIER_END,
        ]
        super().__init__(
            state_map,
            symbol_map,
            bracket_states,
            bracket_ends,
            result_map,
            no_store,
            Token.START
        )
        self.process = self._start

    # Value transforms.
    def _tf_boolean(self, value: str) -> bool:
        if value == 'T':
            return True
        return False

    def _tf_maps(self, value:str) -> tuple[str, dict]:
        mlexer = maps.Lexer()
        lexed = mlexer.lex(value)
        mparser = maps.Parser()
        parsed = mparser.parse(lexed)
        return parsed

    def _tf_number(self, value: str) -> int:
        return int(value)

    def _tf_pool(self, value: str) -> tuple[int, ...]:
        plexer = PoolLexer()
        lexed = plexer.lex(value)
        gen = (L[1] for L in lexed if L[0] == Token.NUMBER)
        return tuple(gen)                           # type: ignore

    def _tf_qualifier(self, value: str) -> str:
        return value[1:-1]

    # Lexing rules.
    def _as_operator(self, char: str) -> None:
        """Processing an operator."""
        can_follow = [
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.GROUP_OPEN,
            Token.U_POOL_DEGEN_OPERATOR,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _boolean(self, char: str) -> None:
        """Processing a boolean."""
        can_follow = [
            Token.CHOICE_OPERATOR,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _choice_operator(self, char: str) -> None:
        """Processing a choice operator."""
        can_follow = [
            Token.QUALIFIER,
            Token.QUALIFIER_DELIMITER,
            Token.CHOICE_OPTIONS,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _comparison_operator(self, char: str) -> None:
        """Processing a comparison operator."""
        can_follow = [
            Token.GROUP_OPEN,
            Token.NEGATIVE_SIGN,
            Token.NUMBER,
            Token.U_POOL_DEGEN_OPERATOR,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _dice_operator(self, char: str) -> None:
        """Processing an operator."""
        can_follow = [
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.GROUP_OPEN,
            Token.U_POOL_DEGEN_OPERATOR,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _ex_operator(self, char: str) -> None:
        """Processing an operator."""
        can_follow = [
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.GROUP_OPEN,
            Token.U_POOL_DEGEN_OPERATOR,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _group_close(self, char: str) -> None:
        """Processing a close group token."""
        can_follow = [
            Token.AS_OPERATOR,
            Token.MD_OPERATOR,
            Token.EX_OPERATOR,
            Token.DICE_OPERATOR,
            Token.GROUP_CLOSE,
            Token.POOL_OPERATOR,
            Token.POOL_GEN_OPERATOR,
            Token.ROLL_DELIMITER,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _group_open(self, char: str) -> None:
        """Processing an open group token."""
        can_follow = [
            Token.GROUP_OPEN,
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.POOL_OPEN,
            Token.U_POOL_DEGEN_OPERATOR,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _map_end(self, char: str) -> None:
        """Processing a choice operator."""
        can_follow = [
            Token.ROLL_DELIMITER,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _map(self, char: str) -> None:
        """Processing a choice operator."""
        self.buffer += char
        if self._is_token_start(Token.MAP_CLOSE, char):
            new_state = Token.MAP_END
            self._change_state(new_state, char)

    def _mapping_operator(self, char: str) -> None:
        can_follow = [
            Token.QUALIFIER,
            Token.QUALIFIER_DELIMITER,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _md_operator(self, char: str) -> None:
        """Processing an operator."""
        can_follow = [
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.GROUP_OPEN,
            Token.U_POOL_DEGEN_OPERATOR,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _number(self, char: str) -> None:
        """Processing a number."""
        can_follow = [
            Token.AS_OPERATOR,
            Token.COMPARISON_OPERATOR,
            Token.DICE_OPERATOR,
            Token.EX_OPERATOR,
            Token.GROUP_CLOSE,
            Token.MAPPING_OPERATOR,
            Token.MD_OPERATOR,
            Token.POOL_GEN_OPERATOR,
            Token.ROLL_DELIMITER,
            Token.WHITESPACE,
        ]

        # Check here if the character is a digit because the checks in
        # Char are currently limited to tokens that no longer than two
        # characters. Check if the state is a number because white
        # space also ends up here, and we want white space to separate
        # numbers.
        if char.isdigit() and self.state == Token.NUMBER:
            self.buffer += char
        else:
            self._check_char(char, can_follow)

    def _options_operator(self, char: str) -> None:
        """Processing an options operator."""
        can_follow = [
            Token.QUALIFIER_DELIMITER,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _pool(self, char: str) -> None:
        """Processing a pool."""
        self.buffer += char
        if self._is_token_start(Token.POOL_CLOSE, char):
            new_state = Token.POOL_END
            self._change_state(new_state, char)

    def _pool_end(self, char: str) -> None:
        """Processing after a pool."""
        can_follow = [
            Token.GROUP_CLOSE,
            Token.POOL_DEGEN_OPERATOR,
            Token.POOL_OPERATOR,
            Token.ROLL_DELIMITER,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _pool_gen_operator(self, char: str) -> None:
        """Processing an pool generation operator."""
        can_follow = [
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.GROUP_OPEN,
            Token.U_POOL_DEGEN_OPERATOR,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _qualifier(self, char: str) -> None:
        """Processing a qualifier."""
        self.buffer += char
        if self._is_token_start(Token.QUALIFIER_DELIMITER, char):
            new_state = Token.QUALIFIER_END
            self._change_state(new_state, char)

    def _qualifier_end(self, char: str) -> None:
        """Process after a qualifier."""
        can_follow = [
            Token.OPTIONS_OPERATOR,
            Token.ROLL_DELIMITER,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _pool_degen_operator(self, char: str) -> None:
        """Processing a pool degeneration operator."""
        can_follow = [
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.GROUP_OPEN,
            Token.U_POOL_DEGEN_OPERATOR,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _pool_operator(self, char: str) -> None:
        """Lex pool operators."""
        can_follow = [
            Token.GROUP_OPEN,
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.U_POOL_DEGEN_OPERATOR,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _roll_delimiter(self, char: str) -> None:
        """Lex roll delimiters."""
        can_follow = [
            Token.MAP_OPEN,
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.BOOLEAN,
            Token.GROUP_OPEN,
            Token.POOL_OPEN,
            Token.U_POOL_DEGEN_OPERATOR,
            Token.QUALIFIER_DELIMITER,
            Token.QUALIFIER,
            Token.BOOLEAN,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _start(self, char: str) -> None:
        """The starting state."""
        if self.tokens:
            self.tokens = []
        self._roll_delimiter(char)

    def _u_pool_degen_operator(self, char: str) -> None:
        """Processing a unary pool degeneration operator."""
        can_follow = [
            Token.GROUP_OPEN,
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.POOL_OPEN,
            Token.U_POOL_DEGEN_OPERATOR,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)


class PoolLexer(BaseLexer):
    def __init__(self) -> None:
        state_map: dict[BaseToken, Callable] = {
            Token.NUMBER: self._number,
            Token.MEMBER_DELIMITER: self._member_delimiter,
            Token.POOL: self._pool,
            Token.POOL_CLOSE: self._pool_close,
            Token.START: self._start,
            Token.WHITESPACE: self._whitespace,
            Token.END: self._start,
        }
        symbol_map: dict[BaseToken, list[str]] = symbols
        bracket_states: dict[BaseToken, BaseToken] = {
            Token.NEGATIVE_SIGN: Token.NUMBER,
            Token.QUALIFIER_DELIMITER: Token.QUALIFIER,
            Token.POOL_OPEN: Token.POOL,
        }
        bracket_ends: dict[BaseToken, BaseToken] = {}
        result_map: dict[BaseToken, Callable] = {
            Token.NUMBER: self._tf_number,
        }
        no_store: list[BaseToken] = [
            Token.POOL_OPEN,
            Token.POOL_CLOSE,
            Token.WHITESPACE,
        ]
        super().__init__(
            state_map,
            symbol_map,
            bracket_states,
            bracket_ends,
            result_map,
            no_store,
            Token.START
        )
        self.process = self._start

    # Value transforms.
    def _tf_number(self, value: str) -> int:
        return int(value)

    # Lexing rules.
    def _member_delimiter(self, char: str) -> None:
        """Lex a member delimiter."""
        can_follow = [
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _number(self, char: str) -> None:
        """Lex a member."""
        can_follow = [
            Token.MEMBER_DELIMITER,
            Token.POOL_CLOSE,
            Token.WHITESPACE,
        ]

        # Check here if the character is a digit because the checks in
        # Char are currently limited to tokens that no longer than two
        # characters. Check if the state is a number because white
        # space also ends up here, and we want white space to separate
        # numbers.
        if char.isdigit() and self.state == Token.NUMBER:
            self.buffer += char
        else:
            self._check_char(char, can_follow)

    def _pool_close(self, char: str) -> None:
        """Lex a pool close."""
        msg = '[ cannot follow a ]'
        raise ValueError(msg)

    def _pool(self, char: str) -> None:
        """Lex a pool open."""
        can_follow = [
            Token.NUMBER,
            Token.NEGATIVE_SIGN,
            Token.WHITESPACE,
        ]
        self._check_char(char, can_follow)

    def _start(self, char: str) -> None:
        """Start lexing the string."""
        can_follow = [
            Token.POOL_OPEN,
        ]
        self._check_char(char, can_follow)



"""
base
~~~~

Base classes for the yadr package.
"""
from abc import ABC, abstractmethod
from typing import Callable, Optional

from yadr.model import BaseToken, CompoundResult, Result, Token, TokenInfo


# Utility functions.
def _mutable(value, type_=list):
    """Return an empty mutable type to avoid bugs where you put a
    mutable in the signature.
    """
    if not value:
        value = type_()
    return value


# Base classes.
class BaseLexer(ABC):
    """An abstract base class for building lexers.

    :class:BaseLexer lexers are state machines used for translating
    a text string into tokens for parsing. It accomplishes this by
    processing the string one character at a time, allowing the
    current state of the lexer to determine whether the character is
    legal and what should be done with it.


    State
    -----
    The current state of the lexer is determined by the value of
    the `state` attribute of the lexer. Its value will be a member
    of the enumeration used to define the tokens that exist within
    the language. This state is used to define the rule used to
    process the next character in the string.

    Characters that do not cause the state of the lexer to change
    should be appended to the end of the `buffer` attribute of
    the lexer.


    State Change
    ------------
    When the lexer encounters a character that represents the end of
    the previous token and the start of a new token, the state of
    the lexer changes. The specific details can vary based on the
    current state of the lexer, but by default the following occurs
    when the state is changed:

    *   A new TokenInfo object is created that contains the current
        state of the lexer and the current value of the `buffer`
        attribute of the lexer.
    *   That TokenInfo object is appended to the `tokens` attribute
        of the lexer.
    *   The `buffer` of the lexer is cleared.
    *   The `state` of the lexer is changed to the new state.
    *   The `process()` method is changed to the process method for
        the new state.


    `BaseLexer.process()` and Processing Methods
    --------------------------------------------
    The `process()` method of a BaseLexer subclass should not be
    defined. Instead the name should be assigned to a "processing"
    method specific to the current state of the lexer. By
    convention, the names of these method starts with an
    underscore, which is followed by the name of the state in
    lowercase letters. So the processing method for the state::

        Token.GROUP_OPEN

    would be::

        _group_open

    The signature for processing methods are::

        (self, char: str) -> None

    where `char` is the character being processed.

    While specific tokens may require different behavior, in general
    a processing method does two things:

    *   Define a list of states that are allowed to follow the
        current state within the syntax being lexed.
    *   pass that list and the character to the internal `_check_char()`
        method that handles the actual processing.

    The end result of calling a processing method is usually that
    the characters in the string that make up the symbol for the
    current state are stored in a "TokenInfo" tuple, which consists
    of the token representing the state and the characters of the
    symbol. These tokens will then be used by the parser to
    execute the command contained in the string.


    The State Map
    -------------
    To determine the correct processing method to use for a state, the
    lexer needs to have a mapping that defines the method for the state.
    This dictionary is the "state map." The tokens for the state are
    the keys, and the processing method for that state is the value
    for the key. This dictionary is passed into the BaseLexer as the
    `state_map` parameter when the lexer is initialized.


    The Symbol Map
    --------------
    A BaseLexer uses a "symbol map" to associate characters in the
    string to a state. The symbol map is a dictionary. The keys are
    the tokens from the enumeration that defines state. The values
    are a list of the strings that are allowed in that state. For
    example, if you have a token named "MULDIV" that is the state for
    multiplication and division operators, the symbol map might look
    like::

        >>> state_map = {
        >>>     Token.MULDIV: ['*', '/'],
        >>> }

    The symbol map is passed to the `symbol_map` parameter when the
    lexer is initialized.


    Bracketing
    ----------
    Instead of running each character through `_check_char()`, it is
    possible for a processing method to instead "bracket" characters
    until a specific character is reached. For example, characters
    after a quotation mark can be collected as a substring until
    the lexer hits another quotation mark.

    Why do this? The main use for this is to turn the bracketed
    substring into a single token, rather than three tokens: the
    opening bracket/delimiter, the content of the bracket, and the
    closing brack/delimiter.

    To expand on the quotation marks example above, let's characters
    surrounded by quotation marks to belong to a token called
    "QUALIFIER". We have the following enumeration of states and
    a `symbol_map` that defines which characters belong to which
    states::

        >>> from enum import auto, Enum
        >>> class Token(Enum):
        >>>     QUALIFIER = auto()
        >>>     DELIM = auto()
        >>>     QUALIFIER_END = auto()
        >>>
        >>> symbol_map = {
        >>>     Token.QUALIFIER: '',
        >>>     Token.DELIM: '"',
        >>>     QUALIFIER_END: '',
        >>> }

    The string we want to lex is::

        >>> text = '"spam"'

    Without a bracket state, you'd end up with a token list that
    would look like the following, assuming the logic for the
    QUALIFIER state is written to accept alphabetical characters
    as valid for qualifiers::

        >>> (
        >>>     (Token.DELIM, '"'),
        >>>     (Token.QUALIFER, 'spam'),
        >>>     (Token.DELIM, '"'),
        >>> )

    That's probably fine, but the delimiter tokens don't really
    do anything at this point. They were just there to set out
    the qualifier in the string. So, you can have them excluded
    from the token list like the following by using bracketing::

        >>> (
        >>>     (Token.QUALIFER, 'spam'),
        >>> )

    The real power here comes from combining with a result map
    to send the bracketed content of to a different lexer and
    parser, which allows syntaxes to be nested within each other.


    Bracket States
    --------------
    To have a processing method bracket, you need to associate
    the state for the opening bracket or delimiter with a
    processing method that handles the bracketing in a dictionary
    passed to the `bracket_states` parameter when initializing the
    BaseLexer. The `bracket_states` dictionary for the above example
    would look like this:

        >>> bracket_state = {
        >>>     Token.DELIM: Token.QUALIFIER,
        >>> }


    Bracket Ends
    ------------
    Because a bracket state hides the closing bracket or delimiter
    from the lexer, you need a different way to handle the state
    after a bracket state. This is handled by a standard processing
    method. By convention the name of this method is an underscore
    followed by the name of the bracket state followed by an
    underscore and then the word "end". For our example it would be::

        _qualifier_end

    This state needs to have a state token assigned for it. In our
    example that is the `Token.QUALIFIER_END` token.

    This end state then needs to be linked to the bracket state in
    a dictionary that is passed to the `bracket_ends` parameter
    when initializing the lexer. In the example, the `bracket_ends`
    dictionary would look like::

        bracket_ends = {
            Token.QUALIFIER: Token.QUALIFIER_END,
        }


    No Store
    --------
    Some states, like the initial state, bracket end states, and
    white space, shouldn't be stored as tokens. These are defined
    by the "no store" list, which is passed to the `no_store`
    parameter when the lexer is initialized. For the example above,
    the no store list could look like::

        >>> no_store = [
        >>>     Token.QUALIFIER_END,
        >>> ]


    Result Transformations
    ----------------------
    By default, a BaseLexer stores the symbols for the token as a
    string in the TokenInfo. This behavior can be changed with a
    "result transformation" method. By convention the name of a
    result transformation starts with an underscore, the letters "tf",
    an underscore, and the name of the state they affect in all lower
    case. So the name of a result transformation method for the
    `Token.NUMBER` state would be::

        _tf_number

    Result transformations have the following signature::

        (self, value:str) -> <type_of_the_transformed_value>

    .. warning:
        The return type of the result transformation method needs
        to be added to the types allowed for TokenInfo. This adds
        complexity that has downstream affects on the parser.

    In the case of something like `Token.NUMBER` the transformation
    can be very simple, coercing a string to an integer. However,
    more complex transformations are possible, such as sending
    bracketed symbols to a different lexer and parser to allow syntax
    nesting.


    Result Map
    ----------
    In order to link the result transformation methods to a state,
    a BaseLexer needs a "result map". The result map is a dictionary.
    The keys are the states where the transforms are used. The values
    are the result transformation methods to use for that state. For
    example, the result map for a lexer that transforms numbers and
    qualifiers might look like::

        >>> result_map = {
        >>>     Token.NUMBER: _tf_number,
        >>>     Token.QUALIFIER: _qualifier,
        >>> }

    The result map is passed to the `result_map` parameter when the
    BaseLexer is initialized.
    """
    def __init__(self,
                 state_map: dict[BaseToken, Callable],
                 symbol_map: dict[BaseToken, list[str]],
                 bracket_states: Optional[dict[BaseToken, BaseToken]] = None,
                 bracket_ends: Optional[dict[BaseToken, BaseToken]] = None,
                 result_map: Optional[dict[BaseToken, Callable]] = None,
                 no_store: Optional[list[BaseToken]] = None,
                 init_state: BaseToken = Token.START) -> None:
        """Initialize an instance of :class:Lexer.

        :param state_map: A dictionary mapping the state of the lexer
            to the processing method for that state.
        :param symbol_map: A dictionary mapping states of the lexer to
            characters that could occur within the text being lexed.
        :param bracket_states: (Optional.) A dictionary mapping opening
            or delimiting states to a state that collects characters
            within the brackets or delimiters to send to a more
            specific lexer.
        :param bracket_ends: (Optional.) A dictionary mapping bracket
            states to a state that processes characters after the end
            of the bracket state.
        :param result_map: (Optional.) A dictionary mapping states
            to a result transformation method to transform the data
            in the lexed string before storing it in as a token
            value.
        :param no_store: (Optional.) A list of states that should not
            be stored as tokens.
        :param init_state: (Optional.) The initial state of the lexer.
            It defaults to :class:Token.START.
        :return: None.
        :rtype: NoneType
        """
        self.state_map = state_map
        self.symbol_map = symbol_map
        self.bracket_states = _mutable(bracket_states, dict)
        self.bracket_ends = _mutable(bracket_ends, dict)
        self.result_map = _mutable(result_map, dict)
        self.no_store = _mutable(no_store)
        self.init_state = init_state
        self.state = init_state

        self.process = self._start
        self.buffer = ''
        self.tokens: list[TokenInfo] = []

    # Public methods.
    def lex(self, yadn: str) -> tuple[TokenInfo, ...]:
        """Lex a dice notation string.

        :param yadn: A string of YADN representing a die roll. For
            details on YADN, see here: YADN_
        :return: A :class:tuple object.
        :rtype: tuple

        .. _YADN: https://github.com/pji/yadr/blob/main/docs/dice_notation.rst
        """
        for char in yadn:
            self.process(char)
        else:
            self._change_state(self.init_state, '')
        return tuple(self.tokens)

    # Private operation method.
    def _is_token_start(self, token: BaseToken, char: str) -> bool:
        """Is the given character the start of a new token."""
        valid = {s[0] for s in self.symbol_map[token]}
        return char in valid

    def _is_token_still(self, char: str) -> bool:
        """Is the given character still a part of the current token."""
        index = len(self.buffer)
        tokens = [t for t in self.symbol_map[self.state] if len(t) > index]
        if tokens:
            valid = {s[index] for s in tokens}
            return char in valid
        return False

    def _cannot_follow(self, char: str) -> None:
        """The character is not allowed by the current state."""
        state = self.state.name
        if state == 'WHITESPACE' and self.tokens:
            state = self.tokens[-1][0].name
        elif state == 'WHITESPACE':
            state = 'START'
        if state == 'QUALIFIER_END':
            state = 'QUALIFIER'

        if state == 'START':
            msg = f'Cannot start with {char}.'
        else:
            article = 'a'
            if state[0] in 'AEIOU':
                article = 'an'
            msg = f'{char} cannot follow {article} {state}.'

        raise ValueError(msg)

    def _change_state(self, new_state: BaseToken, char: str) -> None:
        """Terminate the previous token and start a new one."""
        # Terminate and store the old token.
        if self.state not in self.no_store:
            value: Result | CompoundResult = self.buffer
            if self.state in self.result_map:
                transform = self.result_map[self.state]
                value = transform(value)
            token_info = (self.state, value)
            self.tokens.append(token_info)

        # Set new state.
        self.buffer = char
        self.state = new_state
        self.process = self.state_map[new_state]

    def _check_char(self, char: str, can_follow: list) -> None:
        """Determine how to process a character."""
        new_state: Optional[BaseToken] = None

        # If the character doesn't change the state, add it to the
        # buffer and stop processing.
        if self._is_token_still(char):
            self.buffer += char
            return None

        # Check to see if the character starts a token that is allowed
        # to follow the current token. Stop looking once you find one.
        for token in can_follow:
            if self._is_token_start(token, char):
                new_state = token
                break

        # If not, throw an exception. Since whitespace isn't a token in
        # YADN, an exception saying a character can't follow WHITESPACE
        # isn't useful. Therefore handle that case by looking at the
        # last stored token.
        else:
            self._cannot_follow(char)

        # Some tokens start a state that doesn't match the token.
        if new_state in self.bracket_states:
            new_state = self.bracket_states[new_state]

        # If the state changed, change the state.
        if new_state:
            self._change_state(new_state, char)

    # Lexing rules.
    @abstractmethod
    def _start(self, char: str) -> None:
        """An abstract method for the processing method used for the
        initial state of the lexer.

        :param char: The character currently being lexed.
        :return: None
        :rtype: NoneType
        """
        # The tokens that are allowed to follow the current state.
        can_follow: list[BaseToken] = []

        # Check to see if the current character causes the lexer
        # to change state.
        self._check_char(char, can_follow)

    def _whitespace(self, char: str) -> None:
        """Lex white space."""
        if char.isspace():
            return None
        prev_state = self.init_state
        if self.tokens:
            prev_state = self.tokens[-1][0]
        if prev_state in self.bracket_ends:
            prev_state = self.bracket_ends[prev_state]
        process = self.state_map[prev_state]
        process(char)



"""
yadr
~~~~

The core of the yadn package.
"""
from argparse import ArgumentParser
from pathlib import Path

from yadr.encode import Encoder
from yadr.lex import Lexer
from yadr.model import CompoundResult, Result
from yadr.parser import dice_map, parse


# Default dice mappings.
DICE_MAP_LOCATION = 'yadr/data/dice_maps.yadn'


# Execute YADN.
def roll(yadn: str, yadn_out: bool = False) -> str | Result | CompoundResult:
    """Execute a string of YADN to roll dice.

    :param yadn: A string of YADN that defines the die roll to execute.
        YADN is described here: YADN_
    :param yadn_out: (Optional.) Whether the output should be in native
        Python objects or YADN notation. The default is native Python
        objects.
    :return: The result depends on the details of the die roll, but
        could be a :class:int, :class:tuple, or :class:str object.
    :rtype: int or tuple or str

    .. _YADN: https://github.com/pji/yadr/blob/main/docs/dice_notation.rst

    Usage::

        >>> import yadr
        >>>
        >>> yadr.roll('3d6')                        # doctest: +SKIP
        16

    The specific result will depend on the YADN being executed. In the
    example above, it will be an integer in the range of three to
    eighteen that is created by generating three random integers in the
    range of one to six.
    """
    def roll_dice(yadn: str, yadn_out: bool) -> str | Result | CompoundResult:
        lexer = Lexer()
        tokens = lexer.lex(yadn)
        result: Result | CompoundResult = parse(tokens)
        if yadn_out:
            encoder = Encoder()
            result = encoder.encode(result)
        return result

    default_maps = read_file(DICE_MAP_LOCATION)
    _ = roll_dice(default_maps, False)
    return roll_dice(yadn, yadn_out)


# Utility.
def read_file(loc: str | Path) -> str:
    """Read test from a file.

    :param loc: The file system location of the file.
    :return: A :class:str object.
    :rtype: str
    """
    path = Path(loc)
    with open(path) as fh:
        contents = fh.read()
    return contents


# Command parsing.
def add_dice_map(loc: str) -> None:
    """Load the dice-maps in the given file into memory.

    :param loc: The location of the file of dice mappings to load.
    :return: None.
    :rtype: NoneType
    """
    yadn = read_file(loc)
    _ = roll(yadn)


def list_dice_maps() -> str:
    """Get the list of the currently loaded dice maps.

    :return: A :class:str object.
    :rtype: str
    """
    add_dice_map(DICE_MAP_LOCATION)
    maps_ = '\n'.join(dice_map)
    return maps_


def parse_cli() -> None:
    """Parse command line options."""
    # Stand up the parser.
    p = ArgumentParser(
        description='Execute YADN syntax to roll dice.',
        prog='yadr'
    )

    # Define the command line arguments.
    p.add_argument(
        'yadn',
        help='A string of YADN describing the die roll.',
        action='store',
        nargs='?',
        type=str
    )
    p.add_argument(
        '--list_dice_maps', '-l',
        help='List the names of the loaded dice maps.',
        action='store_true'
    )
    p.add_argument(
        '--add_dice_map', '-m',
        help='Load the dice mappings at the given file location.',
        nargs=1,
        action='store',
        type=str
    )

    # Parse and execute the command.
    args = p.parse_args()
    result = ''
    if args.add_dice_map:
        add_dice_map(args.add_dice_map[0])
    elif args.list_dice_maps:
        result = list_dice_maps()
    if args.yadn:
        raw_result = roll(args.yadn, True)
        result = str(raw_result)
    print(result)



"""
operator
~~~~~~~~

Operators for handling the dice part of dice notation.
"""
import random
import operator
from typing import Sequence


# Choice operators.
def choice_options(a: str, b: str) -> tuple[str, str]:
    """Create the options for a choice."""
    return (a, b)


def choice(boolean: bool, options: tuple[str, str]) -> str:
    """Make a choice."""
    result = options[0]
    if not boolean:
        result = options[1]
    return result


# Pool generation operator.
def dice_pool(num: int, size: int) -> tuple[int, ...]:
    """Roll a die pool."""
    return tuple(random.randint(1, size) for _ in range(num))


def exploding_pool(num: int, size: int) -> tuple[int, ...]:
    """Roll a die pool."""
    pool: Sequence[int] = dice_pool(num, size)
    pool = [_explode(n, size) for n in pool]
    return tuple(pool)


# Dice operators.
def concat(num: int, size: int) -> int:
    """Concatenate the least significant digits."""
    base = 10
    pool = dice_pool(num, size)
    pool = pool_modulo(pool, base)
    return pool_concatenate(pool)


def die(num: int, size: int) -> int:
    """Roll a number of same-sized dice and return the result."""
    pool = dice_pool(num, size)
    return sum(pool)


def exploding_die(num: int, size: int) -> int:
    """Roll a number of exploding same-sized dice."""
    return sum(exploding_pool(num, size))


def keep_high_die(num: int, size: int) -> int:
    """Roll a number of dice and keep the highest."""
    pool = dice_pool(num, size)
    return max(pool)


def keep_low_die(num: int, size: int) -> int:
    """Roll a number of dice and keep the lowest."""
    pool = dice_pool(num, size)
    return min(pool)


def wild_die(num: int, size: int) -> int:
    """Roll a number of same-sized dice and return the result."""
    wild = exploding_pool(1, size)
    regular = dice_pool(num - 1, size)
    if wild[0] == 1:
        return 0
    return sum((sum(wild), sum(regular)))


# Pool operators.
def pool_cap(pool: Sequence[int], cap: int) -> tuple[int, ...]:
    """Cap the maximum value in a pool."""
    result = []
    for value in pool:
        if value > cap:
            value = cap
        result.append(value)
    return tuple(result)


def pool_floor(pool: Sequence[int], floor: int) -> tuple[int, ...]:
    """Floor the minimum value in a pool."""
    result = []
    for value in pool:
        if value < floor:
            value = floor
        result.append(value)
    return tuple(result)


def pool_keep_above(pool: Sequence[int], floor: int) -> tuple[int, ...]:
    return tuple(n for n in pool if n >= floor)


def pool_keep_below(pool: Sequence[int], ceiling: int) -> tuple[int, ...]:
    return tuple(n for n in pool if n <= ceiling)


def pool_keep_high(pool: Sequence[int], keep: int) -> tuple[int, ...]:
    """Keep a number of the highest dice."""
    pool = list(pool)
    remove = len(pool) - keep
    for _ in range(remove):
        low_value = max(pool)
        low_index = 0
        for i, n in enumerate(pool):
            if n < low_value:
                low_value = n
                low_index = i
        pool.pop(low_index)
    return tuple(pool)


def pool_keep_low(pool: Sequence[int], keep: int) -> tuple[int, ...]:
    """Keep a number of the lowest dice."""
    pool = list(pool)
    remove = len(pool) - keep
    for _ in range(remove):
        high_value = min(pool)
        high_index = 0
        for i, n in enumerate(pool):
            if n > high_value:
                high_value = n
                high_index = i
        pool.pop(high_index)
    return tuple(pool)


def pool_modulo(pool: Sequence[int], divisor: int) -> tuple[int, ...]:
    """Perform a modulo operation of each member."""
    return tuple(n % divisor for n in pool)


def pool_remove(pool: Sequence[int], cut: int) -> tuple[int, ...]:
    """Remove members of a pool of the given value."""
    return tuple(n for n in pool if n != cut)


# Pool degeneration operators.
def pool_concatenate(pool: Sequence[int]) -> int:
    """Count the dice in the pool."""
    str_value = ''.join((str(m) for m in pool))
    return int(str_value)


def pool_count(pool: Sequence[int]) -> int:
    """Count the dice in the pool."""
    return len(pool)


def pool_sum(pool: Sequence[int]) -> int:
    """Sum the dice in the pool."""
    return sum(pool)


def count_successes(pool: Sequence[int], target: int) -> int:
    """Count the number of successes in the pool."""
    pool = pool_keep_above(pool, target)
    return len(pool)


def count_successes_with_botch(pool: Sequence[int], target: int) -> int:
    botches = len([n for n in pool if n == 1])
    pool = pool_keep_above(pool, target)
    return len(pool) - botches


# Utility functions.
def _seed(seed: int | str | bytes) -> None:
    """Seed the random number generator for testing purposes."""
    if isinstance(seed, str):
        seed = bytes(seed, encoding='utf_8')
    if isinstance(seed, bytes):
        seed = int.from_bytes(seed, 'little')
    random.seed(seed)


def _explode(value: int, size: int) -> int:
    if value == size:
        explode_value = random.randint(1, size)
        value += _explode(explode_value, size)
    return value


# Registration.
ops_by_symbol = {
    # Choice operator.
    '?': choice,

    # Comparison operators.
    '>': operator.gt,
    '<': operator.le,
    '>=': operator.ge,
    '<=': operator.le,
    '==': operator.eq,
    '!=': operator.ne,

    # Dice operators.
    'd': die,
    'd!': exploding_die,
    'dc': concat,
    'dh': keep_high_die,
    'dl': keep_low_die,
    'dw': wild_die,

    # Operators.
    '^': operator.pow,

    # Multiplication/division operators.
    '*': operator.mul,
    '/': operator.floordiv,
    '%': operator.mod,

    # Addition/subtraction operators.
    '+': operator.add,
    '-': operator.sub,

    # Pool degeneration operators.
    'nb': count_successes_with_botch,
    'ns': count_successes,

    # Pool generation operators.
    'g': dice_pool,
    'g!': exploding_pool,

    # Pool operators.
    'pa': pool_keep_above,
    'pb': pool_keep_below,
    'pc': pool_cap,
    'pf': pool_floor,
    'ph': pool_keep_high,
    'pl': pool_keep_low,
    'pr': pool_remove,
    'p%': pool_modulo,

    # Options operator.
    ':': choice_options,

    # Unary pool degeneration operators.
    'C': pool_concatenate,
    'N': pool_count,
    'S': pool_sum,
}



"""
parse
~~~~~

Parse dice notation.
"""
from functools import wraps
import operator
from typing import Callable, Generic, Optional, Sequence, TypeVar

from yadr import operator as yo
from yadr.model import (
    CompoundResult,
    Result,
    Token,
    TokenInfo,
    op_tokens,
    id_tokens
)


# The dice map.
# This needs to not be a global value, but it will require a very large
# change to this module to get that to work. This will work for now.
dice_map: dict[str, dict] = {}


# Parser specific operations.
def map_result(result: int | tuple[int, ...],
               key: str) -> str | tuple[str, ...]:
    """Map a roll result to a dice map."""
    if isinstance(result, int):
        return dice_map[key][result]
    new_result = [map_result(n, key) for n in result]
    str_result = tuple(str(item) for item in new_result)
    return str_result


# Utility classes and functions.
class Tree:
    """A binary tree."""
    def __init__(self,
                 kind: Token,
                 value: int | str,
                 left: Optional['Tree'] = None,
                 right: Optional['Tree'] = None) -> None:
        self.kind = kind
        self.value = value
        self.left = left
        self.right = right

    def __repr__(self):
        name = self.__class__.__name__
        return f'{name}(kind={self.kind}, value={self.value})'

    def compute(self):
        if self.kind in id_tokens:
            return self.value
        left = self.left.compute()
        right = self.right.compute()
        if self.kind in op_tokens:
            ops_by_symbol = yo.ops_by_symbol
            ops_by_symbol['m'] = map_result
            op = ops_by_symbol[self.value]
        else:
            msg = f'Unknown token {self.kind}'
            raise TypeError(msg)
        return op(left, right)


class Unary(Tree):
    """A unary tree."""
    def __init__(self,
                 kind: Token,
                 value: int | str,
                 child: Optional['Tree'] = None) -> None:
        self.kind = kind
        self.value = value
        self.child = child

    def compute(self):
        if self.kind in id_tokens:
            return self.value
        child = self.child.compute()
        if self.kind in op_tokens:
            op = yo.ops_by_symbol[self.value]
        return op(child)


def next_rule(next_rule: Callable) -> Callable:
    """A decorator for simplifying parsing rules."""
    def outer_wrapper(fn: Callable) -> Callable:
        @wraps(fn)
        def inner_wrapper(*args, **kwargs) -> Callable:
            return fn(next_rule, *args, **kwargs)
        return inner_wrapper
    return outer_wrapper


def u_next_rule(next_rule: Callable) -> Callable:
    """A decorator for simplifying parsing rules."""
    def outer_wrapper(fn: Callable) -> Callable:
        @wraps(fn)
        def inner_wrapper(*args, **kwargs) -> Callable:
            return fn(next_rule, *args, **kwargs)
        return inner_wrapper
    return outer_wrapper


# Parsing initiation.
def parse(tokens: Sequence[TokenInfo]) -> Result | CompoundResult:
    """Parse dice notation tokens."""
    if (Token.ROLL_DELIMITER, ';') not in tokens:
        return _parse_roll(tokens)              # type: ignore

    rolls = []
    while (Token.ROLL_DELIMITER, ';') in tokens:
        index = tokens.index((Token.ROLL_DELIMITER, ';'))
        roll = tokens[0:index]
        rolls.append(roll)
        tokens = tokens[index + 1:]
    else:
        rolls.append(tokens)
    results: Sequence[Result] = []
    for roll in rolls:
        results.append(parse(roll))             # type: ignore
        results = [result for result in results if result is not None]
    if len(results) > 1:
        return CompoundResult(results)
    elif results:
        return results[0]
    return None


def _parse_roll(tokens: Sequence[TokenInfo]) -> Result:
    trees = [Tree(*token) for token in tokens]      # type: ignore
    trees = trees[::-1]
    parsed = last_rule(trees)
    if parsed:
        return parsed.compute()
    return None


# Rule templates.
def _binary_rule(token: Token,
                 next_rule: Callable,
                 trees: list[Tree]) -> Tree:
    """A binary parsing rule."""
    left = next_rule(trees)
    while (trees and trees[-1].kind == token):
        tree = trees.pop()
        tree.left = left
        tree.right = next_rule(trees)
        left = tree
    return left


# Parsing rules.
def groups_and_identity(trees: list[Tree]) -> Tree:
    """Final rule, covering identities, groups, and maps."""
    kind = trees[-1].kind
    value = trees[-1].value
    if kind in id_tokens:
        return trees.pop()
    elif kind == Token.MAP and isinstance(value, tuple):
        name, map_ = value
        global dice_map
        dice_map[name] = map_
        return None
    elif kind == Token.GROUP_OPEN:
        _ = trees.pop()
    else:
        msg = f'Unrecognized token {kind}'
        raise TypeError(msg)
    expression = last_rule(trees)
    if trees[-1].kind == Token.GROUP_CLOSE:
        _ = trees.pop()
    return expression


@next_rule(groups_and_identity)
def pool_gen_operators(next_rule: Callable, trees: list[Tree]):
    """Parse dice operations."""
    return _binary_rule(Token.POOL_GEN_OPERATOR, next_rule, trees)


@next_rule(pool_gen_operators)
def pool_operators(next_rule: Callable, trees: list[Tree]):
    """Parse dice operations."""
    return _binary_rule(Token.POOL_OPERATOR, next_rule, trees)


@u_next_rule(pool_operators)
def u_pool_degen_operators(next_rule: Callable, trees: list[Tree]):
    """Parse dice operations."""
    if trees[-1].kind == Token.U_POOL_DEGEN_OPERATOR:
        tree = trees.pop()
        unary = Unary(tree.kind, tree.value)
        unary.child = next_rule(trees)
        return unary
    return next_rule(trees)


@next_rule(u_pool_degen_operators)
def pool_degen_operators(next_rule: Callable, trees: list[Tree]):
    """Parse dice operations."""
    return _binary_rule(Token.POOL_DEGEN_OPERATOR, next_rule, trees)


@next_rule(pool_degen_operators)
def dice_operators(next_rule: Callable, trees: list[Tree]):
    """Parse dice operations."""
    return _binary_rule(Token.DICE_OPERATOR, next_rule, trees)


@next_rule(dice_operators)
def exponents(next_rule: Callable, trees: list[Tree]):
    """Parse exponents."""
    return _binary_rule(Token.EX_OPERATOR, next_rule, trees)


@next_rule(exponents)
def mul_div(next_rule: Callable, trees: list[Tree]):
    """Parse multiplication and division."""
    return _binary_rule(Token.MD_OPERATOR, next_rule, trees)


@next_rule(mul_div)
def add_sub(next_rule: Callable, trees: list[Tree]):
    """Parse addition and subtraction."""
    return _binary_rule(Token.AS_OPERATOR, next_rule, trees)


@next_rule(add_sub)
def comparison_op(next_rule: Callable, trees: list[Tree]):
    """Parse comparison operator."""
    return _binary_rule(Token.COMPARISON_OPERATOR, next_rule, trees)


@next_rule(comparison_op)
def options_op(next_rule: Callable, trees: list[Tree]):
    """Parse options operator."""
    return _binary_rule(Token.OPTIONS_OPERATOR, next_rule, trees)


@next_rule(options_op)
def choice_op(next_rule: Callable, trees: list[Tree]):
    """Parse options operator."""
    return _binary_rule(Token.CHOICE_OPERATOR, next_rule, trees)


@next_rule(choice_op)
def map_op(next_rule: Callable, trees: list[Tree]):
    """Parse options operator."""
    return _binary_rule(Token.MAPPING_OPERATOR, next_rule, trees)


# Set the last rule in order of operations to make it a little easier
# to update as new operations are added.
last_rule = map_op
